/* cmd.c */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

/* This file contains the cmdread() and cmdparse() functions, which read and
 * parse a command, storing the values into a struct which is then used
 * to adjust a loop.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include "unixsnd.h"

/* This macro handles errors in command lines.  When a command is detected, an
 * error message is written to stderr, along with the erroneous command line.
 * Then it loops back to the beginning of readcmd(), to try the next line.
 */
#define BADCMD(msg)	{\
			   if (ret.loop) loopfree(ret.loop);\
			   fprintf(stderr, "unixsnd: %s: %s\n", msg, cmdline);\
			   return NULL;\
			}

typedef struct macro_s
{
	struct macro_s	*next;	/* some other macro in list */
	char		*name;	/* name of macro */
	char		*text;	/* replacement text of macro */
	char		*back;	/* for stepping through text while expanding */
	struct macro_s 	*from;	/* macro which is invoking this one */
} macro_t;

macro_t	*macros = NULL;

/* Find a macro with a given name */
static macro_t *findmacro(const char *name, int namelen)
{
	macro_t	*mac;

	for (mac = macros; mac; mac = mac->next)
		if (!strncmp(mac->name, name, namelen) && !mac->name[namelen])
			return mac;
	return NULL;
}

/* Define a sound macro */
void cmddefine(const char *name, int namelen, const char *text)
{
	macro_t	*mac;

	/* Is it a previously defined macro? */
	mac = findmacro(name, namelen);
	if (mac)
	{
		/* YES! Old macro */

		/* if currently being expanded, then can't change it */
		if (mac->back)
		{
			fprintf(stderr, "unixsnd: macros aren't allowed to redefine themselves\n");
			return;
		}

		/* replace old definition with new */
		free(mac->text);
		mac->text = strdup(text);
	}
	else
	{
		/* NO! New macro */

		/* allocate a new macro */
		mac = (macro_t *)malloc(sizeof(macro_t));
		mac->next = macros;
		macros = mac;
		mac->name = (char *)malloc(namelen + 1);
		strncpy(mac->name, name, namelen);
		mac->name[namelen] = '\0';
		mac->text = strdup(text);
		mac->back = NULL;
		mac->from = NULL;
	}

	if (verbose)
		printf("macro name=\"%s\"\n      text=\"%s\"\n", mac->name, mac->text);
}


/* Expand macros in a command line.  Returns NULL if failure or the expanded
 * text if success.
 */
static char *expand(char *text)
{
	static char	buf[2000];
	int		used;
	macro_t		*mac, *calling;
	int		namelen;

	/* if no macros, then just return the original text */
	if (!strchr(text, '$'))
		return text;

	/* for each character in text */
	for (mac = NULL, used = 0; *text; )
	{
		/* detect impending doom */
		if (used + 2 > sizeof buf)
		{
			fprintf(stderr, "unixsnd: macro would overflow\n");
			return NULL;
		}

		/* is this a macro? */
		if (*text != '$' || (!isalnum(text[1]) && text[1] != '_' && text[1] != '#'))
		{
			/* NO! literal char, or $$ meaning $ */
			if (*text == '$' && text[1] == '$')
				text++;
			buf[used++] = *text++;
		}
		else
		{
			/* YES! find the end of the name */
			text++;
			for (namelen = 0;
			     isalpha(text[namelen])
			     	|| text[namelen] == '_' || text[namelen] == '#';
			     namelen++)
			{
			}

			/* Try to find a macro for that name */
			calling = findmacro(text, namelen);
			if (!calling)
			{
				fprintf(stderr, "unixsnd: unknown macro: %.*s\n", namelen, text);
				return NULL;
			}

			/* Don't allow recursion */
			if (calling->back)
			{
				fprintf(stderr, "unixsnd: macro would recurse: %.*s\n", namelen, text);
				return NULL;
			}

			/* if zero-length macro, then fail.  Bearing in mind
			 * that macros are normally used to assign sounds to
			 * events, this gives us an easy way to make certain
			 * events be silent.
			 */
			if (!calling->text || !*calling->text)
				return NULL;

			/* start expanding the new macro */
			calling->back = text + namelen;
			calling->from = mac;
			text = calling->text;
			mac = calling;
		}

		/* if reached end of a macro, then resume from previous spot */
		while (!*text && mac)
		{
			text = mac->back;
			mac->back = NULL;
			mac = mac->from;
		}
	}
	buf[used] = '\0';
	return buf;
}


/* Read a command line, and build a command structure from the values given
 * there.  Erroneous commands are skipped.  If "fp" is NULL then it reads from
 * stdin, with the following rules: If EOF is detected, it exits without
 * returning.  If no command is immediately available, then it returns NULL
 * without waiting.
 *
 * When reading from some a file ("fp" not NULL) it returns NULL at EOF.
 */
cmd_t *cmdread(FILE *fp)
{
	cmd_t		*ret;
	char		cmdline[1000];
	fd_set		rfds;
	struct timeval	tv;
	char		*end;

TryAgain:
	if (!fp)
	{
		if (mute)
		{
			if (verbose)
				printf("unixsnd muted\n");
		}
		else if (loopidle() && reverbidle())
		{
			if (verbose)
				printf("unixsnd server idle\n");
		}
		else
		{
			/* if no command is available immediately, then return NULL */
			FD_ZERO(&rfds);
			FD_SET(0, &rfds);
			tv.tv_sec = tv.tv_usec = 0L;
			if (select(1, &rfds, NULL, NULL, &tv) == 0)
				return NULL;
		}
	}

	/* read the command line */
	if (!fgets(cmdline, sizeof cmdline, fp ? fp : stdin))
	{
		if (fp)
			return NULL;
		if (verbose)
			printf("unixsnd server exiting\n");
		exit(0);
	}

	/* strip the terminating '\n' from the line */
	end = strchr(cmdline, '\n');
	if (end)
		*end = '\0';

	/* parse it and return the parsed command */
	ret = cmdparse(cmdline);
	if (!ret)
		goto TryAgain;
	return ret;
}


void cmdscript(char *filename)
{
	FILE	*fp;
	cmd_t	*cmd;

	/* open the file */
	fp = fopen(filename, "r");
	if (!fp)
	{
		perror(filename);
		return;
	}

	/* execute each request from the file */
	while ((cmd = cmdread(fp)) != NULL)
		loopcmd(cmd);

	/* close the file */
	fclose(fp);
}



/* randomly choose one item from a list of alternatives */
static char *choose(const char *list)
{
 static char	buf[80];
	int	alt;
  const char	*scan;
	char	*build;

	/* count the alternatives */
	for (scan = list, alt= 1; *scan; scan++)
		if (*scan == '|')
			alt++;

	/* if only 1 choice, use it */
	if (alt== 1)
	{
		strncpy(buf, list, sizeof buf);
		buf[sizeof buf - 1] = '\0';
		return buf;
	}

	/* else choose one alternative randomly */
	alt = (rand() >> 8) % alt;

	/* find the alternative */
	while (alt > 0)
		if (*list++ == '|')
			alt--;

	/* copy it into the buffer */
	for (build=buf; *list && *list != '|' && build < &buf[sizeof buf - 1]; )
		*build++ = *list++;
	*build = '\0';

	/* return it */
	return buf;
}



/* Parse the fields of a command line, and return it in a struct.  If the
 * line has no fields then return NULL.  (NULL doesn't necessarily indicate
 * an error condition, because some valid commands [e.g, scripts and macro
 * definitions] don't directly produce sounds.  NULL just means that no
 * further processing is required.)
 */
cmd_t *cmdparse(char *cmdline)
{
	static cmd_t	ret;
	char		*cp, *name, *value;
	char		aftervalue;
	int		setleft, setright, setpitch;
	int		setglobal;
	int		firstword;
	int		i;

	/* skip leading whitespace */
	while (*cmdline && isspace(*cmdline))
		cmdline++;

	/* if macro definition, then do that */
	for (cp = cmdline; isalnum(*cp) || *cp == '_' || *cp == '#'; cp++)
	{
	}
	if (cp != cmdline && *cp == ':')
	{
		cmddefine(cmdline, (int)(cp - cmdline), cp + 1);
		return NULL;
	}

	/* Expand any macros in cmdline */
	cmdline = expand(cmdline);
	if (!cmdline)
		return NULL;

	/* if comment or blank line, then simply return NULL */
	if (*cmdline == '#' || !*cmdline)
		return NULL;

	/* if script request, then run the script */
	if (*cmdline == '@')
	{
		cmdscript(cmdline + 1);
		return NULL;
	}

	/* initialize "ret" */
	ret.loop = NULL;
	ret.sound = NULL;
	ret.later = NULL;
	ret.left = ret.right = ret.pitch = 100;
	setleft = setright = setpitch = setglobal = 0;

	/* for each field in the command... */
	for (cp = cmdline, firstword = 1; *cp; firstword = 0)
	{
		/* skip whitespace */
		while (*cp == ' ' || *cp == '\t')
		{
			cp++;
		}

		/* if end of line, then quit (happens if whitespace @ eol) */
		if (!*cp)
			break;

		/* if ';' then store the remainder of the line for later */
		if (*cp == ';')
		{
			cp++;
			ret.later = strdup(cp);
			break;
		}

		/* parse the field name */
		name = cp;
		while (*cp && strchr("=; \t", *cp) == NULL)
		{
			cp++;
		}

		/* parse the field value */
		if (*cp == '=')
		{
			/* skip the '=' */
			cp++;

			/* parse the value */
			for (value = cp; *cp && !isspace(*cp) && *cp != ';'; )
			{
				cp++;
			}
		}
		else /* no '=' given */
		{
			/* assume we got a value, and name is  "sound" */
			value = name;
			name = "sound=";
		}

		/* check for known fields */
		if (!strncmp(name, "loop=", 5))
		{
			aftervalue = *cp;
			*cp = '\0';
			ret.loop = loopfind(value);
			*cp = aftervalue;
		}
		else if (!strncmp(name, "sound=", 6))
		{
			/* find & remember the sound.  If the sound is a list
			 * of possible sounds, then choose one now.
			 */
			aftervalue = *cp;
			*cp = '\0';
			ret.sound = soundfind(choose(value));
			*cp = aftervalue;

			/* Choosing a sound wipes out old pitch & volume,
			 * forcing the new pitch and volume to be altered
			 * as though they were explicitly set.
			 */
			setleft = setright = setpitch = 1;
		}
		else if (!strncmp(name, "left=", 5))
		{
			ret.left = ret.left * atoi(value) / 100;
			setleft = 1;
		}
		else if (!strncmp(name, "right=", 6))
		{
			ret.right = ret.right * atoi(value) / 100;
			setright = 1;
		}
		else if (!strncmp(name, "volume=", 7))
		{
			i = atoi(value);
			ret.left = ret.left * i / 100;
			ret.right = ret.right * i / 100;
			setleft = setright = 1;
		}
		else if (!strncmp(name, "pitch=", 6))
		{
			ret.pitch = ret.pitch * atoi(value) / 100;
			setpitch = 1;
		}
		else if (!strncmp(name, "reverb=", 7))
		{
			reverbinit(value);
			setglobal = 1;
		}
		else if (!strncmp(name, "mute=", 5))
		{
			mute = (*value != 'N' && *value != 'n');
			if (mute)
				loopmute();
			setglobal = 1;
		}
		else
		{
			BADCMD("unknown field name");
		}
	}

	/* Every command needs "loop" or "sound" */
	if (!ret.loop && !ret.sound)
	{
		if (setglobal)
			return NULL;
		else
			BADCMD("missing \"loop\" or \"sound\"");
			/* ... which also returns NULL */
	}

	/* if muted, ignore commands which have no "loop=" */
	if (mute && ret.loop == NULL)
		return NULL;

	/* defaults for everything that wasn't explicitly set */
	if (ret.loop == NULL)
		ret.loop = loopfind(NULL);
	if (ret.sound == NULL)
		ret.sound = ret.loop->sound;
	if (ret.later == NULL)
		ret.later = ret.loop->later;
	if (ret.left < 0 || ret.left > 5000 || !setleft)
		ret.left = ret.loop->left;
	if (ret.right < 0 || ret.right > 5000 || !setright)
		ret.right = ret.loop->right;
	if (ret.pitch < 0 || ret.pitch > 5000 || !setpitch)
		ret.pitch = ret.loop->pitch;

	/* return the command */
	return &ret;
}
