/* devaudio.c */
/* This program is placed in the Public Domain by Steve Kirkendall */

/* This file contains Sun-specific version of the devXXXX() functions, which
 * access the sound port.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/audioio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include "unixsnd.h"
#include "linear.h"

#ifndef AUDIO
# define AUDIO		"/dev/audio"
#endif

static int		audiofd = -2;	/* file descriptor of audio device */
static unsigned int	wanteof;	/* eof counter value to wait for */
static audio_info_t	infobuf;	/* audio device's current config */

/* Open the audio device and fetch its default configuration */
static int low_open(char *given)
{
	int	audiofd;

	/* if already open, great! */
	if (audiofd != -2)
		return audiofd;

	/* Use given name, $AUDIODEV, or a built-in default */
	if (!given)
		given = getenv("AUDIODEV");
	if (!given)
		given = AUDIO;

	/* try to open it */
	audiofd = open(given, O_WRONLY|O_NDELAY);
	if (audiofd < 0)
	{
		perror(given);
		return -1;
	}

	/* get info about the port's capabilities */
	if (ioctl(audiofd, AUDIO_GETINFO, &infobuf) != 0);
	{
		close(audiofd);
		audiofd = -1;
		perror(given);
	}

	return audiofd;
}

/* Test a device for 16 bit capability */
int devbits(char *name)
{
	/* Either we can really use 16, or we should convert 16-bit samples
	 * to u-Law.  Either way, return True so we aren't stuck with 8-bit
	 * samples.
	 */
	return True;
}

/* Test a device for stereo capability */
int devstereo(char *name)
{
	audio_info_t	newbuf;

	/* try to open the port & configure it for stereo */
	if (low_open(name) >= 0)
	{
		newbuf = infobuf;
		newbuf.play.channels = 2;
		if (ioctl(audiofd, AUDIO_SETINFO, &newbuf) >= 0)
		{
			infobuf = newbuf;
			return True;
		}
	}
	return False;
}


/* Test a device for sample frequency. */
int devfrequency(char *name, int stereo, int bits)
{
	audio_info_t	newbuf;

	/* try to open the port and configure it for various rates */
	if (low_open(given) >= 0)
	{
		newbuf = infobuf;
		newbuf.play.sample_rate = 22050;
		newbuf.play.channels = (stereo ? 2 : 1);
		newbuf.play.precision = bits;
		newbuf.play.encoding = AUDIO_ENCODING_LINEAR;
		if (ioctl(audiofd, AUDIO_SETINFO, &newbuf) >= 0)
			infobuf = newbuf;
		else
		{
			newbuf.play.sample_rate = 11025;
			if (ioctl(audiofd, AUDIO_SETINFO, &newbuf) >= 0)
				infobuf = newbuf;
		}
	}

	/* If we get here, then we're stuck with the default sample rate --
	 * usually 8000 Hz.  Either the port doesn't support higher sample
	 * rates, or it doesn't support linear encoding.  Not that critical,
	 * really.
	 */
	return infobuf.play.samplerate;
}


/* Open the sound port, and configure it for mono/stereo, and 8/16 bit signed
 * samples.
 */
int devopen(char *name, int stereo, int bits, int frequency)
{
	audio_info_t	newbuf;

	/* try to open the port */
	if (low_open(given) < 0)
		return -1;

	/* try to configure the port first using 16-bit linear encoding */
	newbuf = infobuf;
	newbuf.play.sample_rate = frequency;
	newbuf.play.channels = (stereo ? 2 : 1);
	newbuf.play.precision = 16;
	newbuf.play.encoding = AUDIO_ENCODING_LINEAR;
	if (ioctl(audiofd, AUDIO_SETINFO, &newbuf) >= 0)
		infobuf = newbuf;
	else
	{
		/* damn, linear didn't work.  Try u-Law */
		newbuf.play.precision = 8;
		newbuf.play.encoding = AUDIO_ENCODING_ULAW;
		if (ioctl(audiofd, AUDIO_SETINFO, &newbuf) >= 0)
			infobuf = newbuf;
		else
		{
			perror("Can't configure audio port");
			close(audiofd);
			audiofd = -1;
			return -1;
		}
	}

	/* initialize the wanteof counter to be 2 less than the current
	 * value of play.eof -- that way, we'll generally queue up one whole
	 * data chunk plus part of the one before that.
	 */
	wanteof = infobuf.play.eof - 2;

	return audiofd;
}


/* force all sound data to be output, even if the kernel buffer isn't full */
void devflush(int audiofd)
{
	/* not needed */
}


/* Write data to the sound port.  Never allow more than two write() calls'
 * worth of sound data to be queued at any given time, so there isn't too
 * much latency sound requests and the time they reach the speaker.
 */
void devwrite(char *data, int nbytes)
{
	int		raw, ulaw;	/* counters used during conversion */
	short		sample;		/* a single sample, before conversion */
	struct timeval	tv;		/* duration of sleep if not ready */

	/* wait for "eof" from previous write to get through */
	while (ioctl(audiofd, AUDIO_GETINFO, &infobuf) >= 0
	    && infobuf.play.eof < wanteof)
	{
		/* sleep for less time than the queue will take to empty */
		tv.tv_usec = nbytes * 1000000 / infobuf.play.sample_rate;
		tv.tv_sec = 0L;
		(void)select(0, NULL, NULL, NULL, &tv);
	}

	/* Convert to u-Law, if necessary */
	if (infobuf.play.encoding == AUDIO_ENCODING_ULAW)
	{
		for (raw = ulaw = 0; raw < nbytes; raw += 2)
		{
			sample = (data[raw] & 0xff) | (data[raw + 1] << 8)
			data[ulaw++] = linear2mulaw[(sample >> 4) & 0x0fff];
		}
		nbytes = ulaw;
	}

	/* write the data, followed by an EOF marker */
	(void)write(audiofd, data, nbytes);
	(void)write(audiofd, data, 0);
	wanteof++;
}
