/* loop.c */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

/* This file contains the loopXXXX() functions, which manage actively-playing
 * sounds.
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "unixsnd.h"

/* linked list of all currently-active sounds */
static loop_t	*loops;

int loopidle(void)
{
	return loops == NULL;
}


/* Find a loop.  If not found, then create it with default values */
loop_t *loopfind(char *name)
{
	loop_t	*ret;

	/* search for the loop in the list of currently-active sounds */
	if (name)
	{
		for (ret = loops; ret; ret = ret->next)
		{
			if (ret->name && !strcmp(ret->name, name))
			{
				if (verbose)
					printf("changing loop \"%s\"\n", name);
				return ret;
			}
		}
	}

	/* allocate a new loop struct */
	if (verbose)
		printf("creating %s loop\n", name ? name : "anonymous");
	ret = (loop_t *)malloc(sizeof(loop_t));
	ret->next = loops;
	loops = ret;
	ret->name = (name ? strdup(name) : NULL);
	ret->left = ret->right = 100;
	ret->pitch = 100;
	ret->sound = NULL;
	ret->later = NULL;
	ret->pos = 0;
	ret->incr = 1;

	/* return the new loop */
	return ret;
}


/* delete a loop */
void loopfree(loop_t *loop)
{
	loop_t	*scan, *lag;
	cmd_t	*cmd = NULL;

	if (verbose)
		printf("deleting %s loop, for %s sound\n",
			loop->name ? loop->name : "*anonymous*",
			loop->sound ? loop->sound->name : "*none*");

	for (scan = loops, lag = NULL; scan != loop; lag = scan, scan = scan->next)
	{
	}
	if (lag)
		lag->next = loop->next;
	else
		loops = loop->next;
	if (loop->name)
		free(loop->name);
	if (loop->later)
	{
		cmd = cmdparse(loop->later);
		free(loop->later);
	}
	free(loop);

	/* if there was a "later" request, do it now */
	if (cmd)
		loopcmd(cmd);
}

/* Adjust a loop as described by a cmd_t structure */
void loopcmd(cmd_t *cmd)
{
	/* if no sound given (or kept from earlier command for this loop),
	 * or if the volume or pitch is reduced to 0, then delete the loop.
	 */
	if (!cmd->sound || cmd->left + cmd->right == 0 || cmd->pitch == 0)
	{
		loopfree(cmd->loop);
		return;
	}

	/* adjust the values of this loop */
	cmd->loop->left = cmd->left;
	cmd->loop->right = cmd->right;
	cmd->loop->pitch = cmd->pitch;
	if (cmd->later != cmd->loop->later)
	{
		if (cmd->loop->later)
			free(cmd->loop->later);
		cmd->loop->later = cmd->later;
	}
	if (cmd->loop->sound != cmd->sound)
	{
		cmd->loop->sound = cmd->sound;
		cmd->loop->pos = 0;
	}
	cmd->loop->incr = (long)cmd->sound->frequency * cmd->pitch * 64
						/ (25 * output_frequency);
	cmd->loop->end = cmd->sound->nsamples * 256L;
}

/* Mix the samples, to produce some sound data.  Return the number of samples
 * generated -- which will always be "count" unless all loops are one-shots
 * that terminate before the end of the data.
 */
int loopmix(mixed_t *sample, int count)
{
	loop_t	*this, *next;
	int	nsamples;
	int	i, j, s;

	/* Zero the sample buffer */
	memset(sample, 0, count * sizeof(mixed_t));
	nsamples = 0;

	/* For each loop... */
	for (this = loops; this; this = next)
	{
		/* Remember which loop follows this one.  We can't depend on
		 * this->next at the end of the loop, because we might free
		 * this loop.
		 */
		next = this->next;

		/* Generate each sample for this loop */
		if (this->sound->stereo)
		{
			/* STEREO VERSION */
			for (i = 0; i < count; i++)
			{
				/* If the current position is past the end of
				 * the sample data, then either delete the loop
				 * (if single-shot) or wrap (if looped).
				 */
				if (this->pos >= this->end)
				{
					if (!this->name)
					{
						loopfree(this);
						break;
					}
					this->pos %= this->end;
				}

				/* Add the weighted sample to the total */
				s = this->sound->sample[this->pos >> 8];
				sample[i].left += s * this->left;
				this->pos++; /* always safe for stereo sounds */
				s = this->sound->sample[this->pos >> 8];
				sample[i].right += s * this->right;

				/* increment the sample position */
				this->pos += this->incr;
			}
		}
		else
		{
			/* MONO VERSION (into stereo channels) */
			for (i = 0; i < count; i++)
			{
				/* If the current position is past the end of
				 * the sample data, then either delete the loop
				 * (if single-shot) or wrap (if looped).
				 */
				if (this->pos >= this->end)
				{
					if (!this->name)
					{
						loopfree(this);
						break;
					}
					this->pos %= this->end;
				}

				/* Add the weighted sample to the total */
				j = this->pos >> 8;
				s = this->sound->sample[j];
				sample[i].left += s * this->left;
				sample[i].right += s * this->right;

				/* increment the sample position */
				this->pos += this->incr;
			}
		}

		/* If this sound was the longest seen so far, then its length
		 * is the number of samples so far.
		 */
		if (i > nsamples)
			nsamples = i;
	}

	/* add reverb */
	reverbsamples(sample, nsamples);

	/* scale and clip the samples */
	for (i = 0; i < nsamples; i++)
	{
		/* left */
		sample[i].left /= 100;
		if (sample[i].left < -25000)
		{
			sample[i].left = -25000 + (sample[i].left + 25000) / 6;
			if (sample[i].left < -32768)
				sample[i].left = -32768;
		}
		else if (sample[i].left > 25000)
		{
			sample[i].left = 25000 + (sample[i].left - 25000) / 6;
			if (sample[i].left > 32767)
				sample[i].left = 32767;
		}

		/* right */
		sample[i].right /= 100;
		if (sample[i].right < -25000)
		{
			sample[i].right = -25000 + (sample[i].right + 25000)/ 6;
			if (sample[i].right < -32768)
				sample[i].right = -32768;
		}
		else if (sample[i].right > 25000)
		{
			sample[i].right = 25000 + (sample[i].right - 25000) / 6;
			if (sample[i].right > 32767)
				sample[i].right = 32767;
		}
	}

	/* return the number of samples generated. */
	return nsamples;
}


/* Cancel any pending non-looped sounds.  This is done at the start of a mute */
void loopmute(void)
{
	loop_t *this, *next;

	for (this = loops; this; this = next)
	{
		next = this->next;
		if (!this->name)
		{
			loopfree(this);
		}
	}
}
