/* unixsnd.c */
/* Placed in public domain by Steve Kirkendall.  Use at your own risk */

/* This contains the main() function, and the main loop, of the sound server */

#include <stdlib.h>
#include <unistd.h> /* for getopt() */
#include "unixsnd.h"
#define SOUNDPATH ":~/.sounds:sounds:/usr/share/sounds/au:/usr/share/sounds/wav"


/* From command-line arguments */
static char	*device = NULL;
static int	stereo = -1;
static int	bits = -1;
int		output_frequency = -1;
int		verbose = -1;
int		mute = 0;
static int	cont = -1;


static int	devfd;
static mixed_t	*mix;
static char	*data;
static int	maxsamples;


/* output a usage message */
static void usage(int exitcode)
{
	fprintf(stderr, "usage: unixsnd [flags] [request]...\n");
	fprintf(stderr, "flags: -v            verbose - give status messages\n");
	fprintf(stderr, "       -q            quiet - don't give status messages\n");
	fprintf(stderr, "       -s            stereo output\n");
	fprintf(stderr, "       -m            mono output\n");
	fprintf(stderr, "       -b bits       output bits (8 or 16)\n");
	fprintf(stderr, "       -r frequency  output sample rate, in Hertz\n");
	fprintf(stderr, "       -p soundpath  add directory to path where sound files reside\n");
	fprintf(stderr, "       -d device     send sound output to the given device\n");
	fprintf(stderr, "       -c            continue after requests from command line\n");
	fprintf(stderr, "       -x            exit after requests from command line\n");
	fprintf(stderr, "Unixsnd plays sounds.  It is normally invoked from an application program\n");
	fprintf(stderr, "which sends it requests via a pipe.  You can also invoke it with requests\n");
	fprintf(stderr, "on the command line; this can be handy when you want to listen to a few\n");
	fprintf(stderr, "sound files.  There are reasonable defaults for all options, so normally\n");
	fprintf(stderr, "you (or an application) will start unixsnd with no command-line flags.\n");
	exit(exitcode);
}


/* Convert samples from a mixed_t array to a series of bytes */
static void writesamples(mixed_t *mixed, int nsamples)
{
	char	*cp;	/* for building 8-bit samples */
	short	*sp;	/* for building 16-bit samples */
	int	i;

	if (bits == 8)
	{
		cp = data;
		for (i = 0; i < nsamples; i++)
		{
			if (stereo)
			{
				*cp++ = mixed[i].left >> 8;
				*cp++ = mixed[i].right >> 8;
			}
			else
			{
				*cp++ = (mixed[i].left + mixed[i].right) >> 9;
			}
		}
		if (stereo)
			i *= 2;
	}
	else /* bits == 16 */
	{
		sp = (short *)data;
		for (i = 0; i < nsamples; i++)
		{
			if (stereo)
			{
				*sp++ = mixed[i].left;
				*sp++ = mixed[i].right;
			}
			else
			{
				*sp++ = (mixed[i].left + mixed[i].right) >> 1;
			}
		}
		if (stereo)
			i *= 4;
		else
			i *= 2;
	}

	/* write the data */
	(void)write(devfd, data, i);
}


void main(int argc, char **argv)
{
	cmd_t	*cmd;
	int	nsamples;
	int	flag;
	int	i;
	char	*value;

	/* check for some special cases for arguments */
	if (argc >= 2 && !strcmp(argv[1], "--help"))
	{
		usage(0);
	}
	else if (argc >= 2 && !strcmp(argv[1], "--version"))
	{
		printf("unixsnd 0.91\n");
		printf("Placed in public domain by the author, Steve Kirkendall\n");
		exit(0);
	}

	/* Try to run at a slightly higher priority than other processes,
	 * to avoid gaps in the sound output.  If this fails, ignore it;
	 * it isn't that critical, and only the superuser can reduce its
	 * niceness.
	 */
	(void)nice(-1);

	/* In order to make the above nice(-1) work, some sites may choose
	 * to install unixsnd as SUID-root.  For their benefit, we will
	 * revert to the normal user ID here, so hacker's can't use root
	 * priviledge to "play" protected files.
	 */
	if (geteuid() != getuid() && seteuid(getuid()) < 0)
	{
		fprintf(stderr, "unixsnd: could not reset security level\n");
		exit(3);
	}

	/* set the default sound path */
	value = getenv("SOUNDPATH");
	if (!value)
		value = SOUNDPATH;
	soundpath(value);

	/* parse arguments */
	while ((flag = getopt(argc, argv, "?vqsmb:r:p:d:cx")) != EOF)
	{
		switch (flag)
		{
		  case '?':
			/* Note that this case may be invoked due to an explicit
			 * "-?" flag on the command line, or due to any unknown
			 * flag.  Either way, the usage message is appropriate.
			 */
			usage(0);
			/*NOTREACHED*/

		  case 'v':
			verbose = 1;
			break;

		  case 'q':
			verbose = 0;
			break;

		  case 's':
			stereo = 1;
			break;

		  case 'm':
			stereo = 0;
			break;

		  case 'b':
			bits = atoi(optarg);
			if (bits != 8 && bits != 16)
			{
				fprintf(stderr, "unixsnd: number of bits (-b) must be 8 or 16\n");
				exit(2);
			}
			break;

		  case 'r':
			output_frequency = atoi(optarg);
			if (output_frequency < 4000 || output_frequency > 48000)
			{
				fprintf(stderr, "unixsnd: output sample rate (-r) must be between 4000 and 48000, and some\n");
				fprintf(stderr, "         sound cards may be limited to an even narrower range.\n");
				exit(2);
			}
			break;

		  case 'p':
			soundpath(optarg);
			break;

		  case 'd':
			device = optarg;
			break;

		  case 'c':
			cont = 1;
			break;

		  case 'x':
			cont = 0;
			break;
		}
	}

	/* anything not explicitly set via arguments should use defaults */
	if (cont == -1)
		cont = (optind == argc);
	if (verbose == -1)
		verbose = cont && isatty(0);
	if (bits == -1)
		bits = devbits(device);
	if (stereo == -1)
		stereo = devstereo(device);
	if (output_frequency == -1)
		output_frequency = devfrequency(device, stereo, bits);

	/* open the sound port */
	devfd = devopen(device, stereo, bits, output_frequency);
	if (devfd < 0)
		exit(2);

	/* allocate the buffers */
	maxsamples = 1024 * (output_frequency + 6000) / 11025;
	mix = (mixed_t *)malloc(maxsamples * sizeof(mixed_t));
	data = (char *)malloc(maxsamples * (stereo ? 2 : 1) * bits / 8);

	if (verbose)
		printf("unixsnd server starting\n");

	/* Handle any requests from command line.  Wait for idleness between
	 * them, so they are played sequentially.
	 */
	for (i = optind; i < argc; i++)
	{
		/* parse & execute the command */
		cmd = cmdparse(argv[i]);
		if (cmd)
			loopcmd(cmd);

		/* Generate sound until idle */
		while ((nsamples = loopmix(mix, maxsamples)) > 0)
		{
			writesamples(mix, nsamples);
		}
	}

	/* if not supposed to continue with requests from stdin, then exit */
	if (!cont)
		exit(0);

	/* main input loop - never exits, because cmdread() exits at EOF */
	for(;;)
	{
		/* handle any pending commands */
		while ((cmd = cmdread(NULL)) != NULL)
		{
			loopcmd(cmd);
		}

		/* output some sample data to the sound device */
		if (!mute)
		{
			nsamples = loopmix(mix, maxsamples);
			if (nsamples == 0)
				nsamples = reverbsilence(mix, maxsamples);
			if (nsamples > 0)
				writesamples(mix, nsamples);
		}

		/* if we became idle, then flush sounds */
		if (mute || (loopidle() && reverbidle()))
			devflush(devfd);
	}

	/*NOTREACHED*/
}
